@@ -3,6 +3,7 @@ import os
 from cryptography.hazmat.backends import default_backend
 from cryptography.hazmat.primitives import hashes
 from cryptography.hazmat.primitives.kdf.pbkdf2 import PBKDF2HMAC
+
 password = input("Password? ").encode("ascii")
 salt = os.urandom(16)
 kdf = PBKDF2HMAC(
@@ -10,7 +11,7 @@ kdf = PBKDF2HMAC(
     length=32,
     salt=salt,
     iterations=100000,
-    backend=default_backend()
+    backend=default_backend(),
 )
 key = base64.urlsafe_b64encode(kdf.derive(password))
 print("Key: ", key)
